#ifndef __GLOBALLINESRENDERER_H
#define __GLOBALLINESRENDERER_H

#include "common.h"
#include "Containers.h"
#include "Vector3.h"


class RenderTexture;

namespace GtpVisibilityPreprocessor {

class Beam;
class Intersectable;
class Preprocessor;
class GlRenderer;

struct VssRayContainer;

class GlobalLinesRenderer
{
public:
	/*GlobalLinesRenderer(RenderTexture *buffer1, 
						RenderTexture *buffer2, 
						Preprocessor *preprocessor,
						GlRenderer *renderer);
*/
	GlobalLinesRenderer(Preprocessor *preprocessor, GlRenderer *renderer);

	~GlobalLinesRenderer();

	void CastGlobalLines(Beam &beam, const int samples);

	void InitGl();

	/** Computes rays from information gained with hw sampling
	*/
	void ComputeRays(Intersectable *sourceObj, VssRayContainer &rays);

	float GetPixelError(int &pvsSize);

	int ComputePvs(ObjectContainer &objects, ObjectContainer &pvs) const;

	void RenderObject(Intersectable *obj);

	
	/** Draws the geometry for one pass.
	*/
	void DrawGeometry();
	void Run();
	void GrabDepthBuffer(float *data, RenderTexture *rt);
	void GrabItemBuffer(unsigned char *data, RenderTexture *rt);
	void ApplyDepthPeeling(VssRayContainer &rays, Beam &beam, const int samples);
	void ExportDepthBuffer();
	void ExportItemBuffer();
	void ProcessDepthBuffer(VssRayContainer &vssRays, const bool oldBufferInitialised, const int pass);
	void DisplayBuffer(const bool isDepth);
	
	bool ClipToViewSpaceBox(const Vector3 &origin,
							const Vector3 &termination,
							Vector3 &clippedOrigin,
							Vector3 &clippedTermination);

	Intersectable *ExtractSamplePoint(float *depthBuffer,
									  unsigned char *itemBuffer,
									  const int x, 
									  const int y,
									  Vector3 &hitPoint,
									  const bool isFrontBuffer) const;

	void Visualize(const VssRayContainer &vssRays);

	int mMaxDepth;
	Vector3 mEyeVec;
	Vector3 mLeftVec;
	Vector3 mUpVec;
	float mFar;
	float mNear;
	float mWidth;

	Vector3 mViewPoint;

	float *mNewDepthBuffer;
	float *mOldDepthBuffer;
	unsigned char *mNewItemBuffer;
	unsigned char *mOldItemBuffer;
	
	RenderTexture *mNewTexture;
	RenderTexture *mOldTexture;

	void ComputeLookAt(const float alpha, 
					   const float beta,
					   Vector3 &eye, 
					   Vector3 &up, 
					   Vector3 left);
protected:

	void SwitchRenderTextures();

	void SetupProjection(const Vector3 &viewPoint, 
						 const Beam &beam);

	

	Preprocessor *mPreprocessor;

	GlRenderer *mRenderer;
};

extern GlobalLinesRenderer *globalLinesRenderer;

};

#endif

